@@ -28,13 +28,17 @@ class MinimalStreamWidget;
 class ChannelWidget : public Gtk::EventBox {
 public:
     ChannelWidget(BaseObjectType* cobject, const Glib::RefPtr<Gtk::Builder>& x);
-    static ChannelWidget* create();
+
+    /* This creates multiple ChannelWidgets based on the given channel map. The
+     * widgets are stored in the caller-provided array. */
+    static void create(MinimalStreamWidget *owner, const pa_channel_map &m, bool can_decibel,
+                       ChannelWidget *widgets[PA_CHANNELS_MAX]);
 
     void setVolume(pa_volume_t volume);
 
     Gtk::Label *channelLabel;
     Gtk::Label *volumeLabel;
-    Gtk::HScale *volumeScale;
+    Gtk::Scale *volumeScale;
 
     int channel;
     MinimalStreamWidget *minimalStreamWidget;
@@ -47,7 +51,10 @@ public:
 
     virtual void set_sensitive(bool enabled);
     virtual void setBaseVolume(pa_volume_t);
+
+private:
+    static ChannelWidget *createOne(MinimalStreamWidget *owner, int channelIndex, pa_channel_position channelPosition,
+                                    bool can_decibel);
 };
 
-
 #endif
