@@ -95,11 +95,11 @@ MainWindow::MainWindow(BaseObjectType* cobject, const Glib::RefPtr<Gtk::Builder>
     x->get_widget("notebook", notebook);
     x->get_widget("showVolumeMetersCheckButton", showVolumeMetersCheckButton);
 
-    cardsVBox->set_reallocate_redraws(true);
-    sourcesVBox->set_reallocate_redraws(true);
-    streamsVBox->set_reallocate_redraws(true);
-    recsVBox->set_reallocate_redraws(true);
-    sinksVBox->set_reallocate_redraws(true);
+    sourcesVBox->signal_size_allocate().connect([this](Gdk::Rectangle _unused){ sourcesVBox->queue_draw(); });
+    cardsVBox->signal_size_allocate().connect([this](Gdk::Rectangle _unused){ cardsVBox->queue_draw(); });
+    streamsVBox->signal_size_allocate().connect([this](Gdk::Rectangle _unused){ streamsVBox->queue_draw(); });
+    recsVBox->signal_size_allocate().connect([this](Gdk::Rectangle _unused){ recsVBox->queue_draw(); });
+    sinksVBox->signal_size_allocate().connect([this](Gdk::Rectangle _unused){ sinksVBox->queue_draw(); });
 
     sinkInputTypeComboBox->set_active((int) showSinkInputType);
     sourceOutputTypeComboBox->set_active((int) showSourceOutputType);
@@ -188,6 +188,7 @@ MainWindow* MainWindow::create(bool maximize) {
     x->add_from_file(GLADE_FILE, "liststore4");
     x->add_from_file(GLADE_FILE, "mainWindow");
     x->get_widget_derived("mainWindow", w);
+    w->get_style_context()->add_class("pavucontrol-window");
     if (w && maximize)
         w->maximize();
     return w;
@@ -221,7 +222,7 @@ bool MainWindow::on_key_press_event(GdkEventKey* event) {
             case GDK_KEY_Q:
             case GDK_KEY_w:
             case GDK_KEY_q:
-                Gtk::Main::quit();
+                close();
                 return true;
         }
     }
@@ -283,17 +284,23 @@ static void set_icon_name_fallback(Gtk::Image *i, const char *name, Gtk::IconSiz
 
     try {
         pixbuf = theme->load_icon(name, width, Gtk::ICON_LOOKUP_GENERIC_FALLBACK | Gtk::ICON_LOOKUP_FORCE_SIZE);
+    } catch (Glib::Error &e) {
+        /* Ignore errors. */
+    }
 
-        if (pixbuf)
-            i->set(pixbuf);
-        else
-            i->set(name);
-    } catch (Gtk::IconThemeError &e) {
-        i->set(name);
-    } catch (Gio::Error &e) {
-        i->set(name);
-    } catch (Gdk::PixbufError &e) {
-        i->set(name);
+    if (!pixbuf) {
+        try {
+            pixbuf = Gdk::Pixbuf::create_from_file(name);
+        } catch (Glib::FileError &e) {
+            /* Ignore errors. */
+        } catch (Gdk::PixbufError &e) {
+            /* Ignore errors. */
+        }
+    }
+
+    if (pixbuf) {
+        pixbuf = pixbuf->scale_simple(width, height, Gdk::INTERP_BILINEAR);
+        i->set(pixbuf);
     }
 }
 
@@ -366,9 +373,13 @@ void MainWindow::updateCard(const pa_card_info &info) {
 
     w->updating = true;
 
+    w->pulse_card_name = info.name;
+
     description = pa_proplist_gets(info.proplist, PA_PROP_DEVICE_DESCRIPTION);
     w->name = description ? description : info.name;
-    w->nameLabel->set_markup(w->name.c_str());
+    gchar *txt;
+    w->nameLabel->set_markup(txt = g_markup_printf_escaped("%s", w->name.c_str()));
+    g_free(txt);
 
     icon = pa_proplist_gets(info.proplist, PA_PROP_DEVICE_ICON_NAME);
     set_icon_name_fallback(w->iconImage, icon ? icon : "audio-card", Gtk::ICON_SIZE_SMALL_TOOLBAR);
@@ -422,7 +433,7 @@ void MainWindow::updateCard(const pa_card_info &info) {
         if (!profileIt->available)
             desc += _(" (unavailable)");
 
-        w->profiles.push_back(std::pair<Glib::ustring,Glib::ustring>(profileIt->name, desc));
+        w->profiles.push_back(std::pair<Glib::ustring, Glib::ustring>(profileIt->name, desc));
     }
 
     w->activeProfile = info.active_profile ? info.active_profile->name : "";
@@ -466,6 +477,73 @@ void MainWindow::updateCard(const pa_card_info &info) {
     w->updating = false;
 }
 
+void MainWindow::updateCardCodecs(const std::string& card_name, const std::unordered_map<std::string, std::string>& codecs) {
+    CardWidget *w = NULL;
+
+    for (auto c : cardWidgets) {
+        if (card_name.compare(c.second->pulse_card_name) == 0)
+            w = c.second;
+    }
+
+    if (!w)
+        return;
+
+    w->updating = true;
+
+    w->codecs.clear();
+
+    /* insert profiles */
+    for (auto e : codecs) {
+        w->codecs.push_back(std::pair<Glib::ustring, Glib::ustring>(e.first, e.second));
+    }
+
+    w->prepareMenu();
+
+    w->updating = false;
+}
+
+void MainWindow::setActiveCodec(const std::string& card_name, const std::string& codec) {
+    CardWidget *w = NULL;
+
+    for (auto c : cardWidgets) {
+        if (card_name.compare(c.second->pulse_card_name) == 0)
+            w = c.second;
+    }
+
+    if (!w)
+        return;
+
+    w->updating = true;
+
+    w->activeCodec = codec;
+
+    w->prepareMenu();
+
+    w->updating = false;
+}
+
+void MainWindow::setCardProfileIsSticky(const std::string& card_name, gboolean profile_is_sticky) {
+    CardWidget *w = NULL;
+
+    for (auto c : cardWidgets) {
+        if (card_name.compare(c.second->pulse_card_name) == 0)
+            w = c.second;
+    }
+
+    if (!w)
+        return;
+
+    w->updating = true;
+
+    /* make sure that profile lock toggle button is visible */
+    w->hasProfileLock = true;
+    w->profileLockToggleButton->set_active(profile_is_sticky);
+
+    w->prepareMenu();
+
+    w->updating = false;
+}
+
 bool MainWindow::updateSink(const pa_sink_info &info) {
     SinkWidget *w;
     bool is_new = false;
@@ -516,7 +594,7 @@ bool MainWindow::updateSink(const pa_sink_info &info) {
 
     w->ports.clear();
     for (std::set<pa_sink_port_info>::iterator i = port_priorities.begin(); i != port_priorities.end(); ++i)
-        w->ports.push_back(std::pair<Glib::ustring,Glib::ustring>(i->name, i->description));
+        w->ports.push_back(std::pair<Glib::ustring, Glib::ustring>(i->name, i->description));
 
     w->activePort = info.active_port ? info.active_port->name : "";
 
@@ -529,13 +607,13 @@ bool MainWindow::updateSink(const pa_sink_info &info) {
     w->setDigital(info.flags & PA_SINK_SET_FORMATS);
 #endif
 
-    w->updating = false;
-
     w->prepareMenu();
 
     if (is_new)
         updateDeviceVisibility();
 
+    w->updating = false;
+
     return is_new;
 }
 
@@ -625,6 +703,7 @@ void MainWindow::createMonitorStreamForSinkInput(SinkInputWidget* w, uint32_t si
 
     if (w->peak) {
         pa_stream_disconnect(w->peak);
+        pa_stream_unref(w->peak);
         w->peak = NULL;
     }
 
@@ -683,7 +762,7 @@ void MainWindow::updateSource(const pa_source_info &info) {
 
     w->ports.clear();
     for (std::set<pa_source_port_info>::iterator i = port_priorities.begin(); i != port_priorities.end(); ++i)
-        w->ports.push_back(std::pair<Glib::ustring,Glib::ustring>(i->name, i->description));
+        w->ports.push_back(std::pair<Glib::ustring, Glib::ustring>(i->name, i->description));
 
     w->activePort = info.active_port ? info.active_port->name : "";
 
@@ -692,12 +771,12 @@ void MainWindow::updateSource(const pa_source_info &info) {
     if (cw != cardWidgets.end())
         updatePorts(w, cw->second->ports);
 
-    w->updating = false;
-
     w->prepareMenu();
 
     if (is_new)
         updateDeviceVisibility();
+
+    w->updating = false;
 }
 
 void MainWindow::setIconFromProplist(Gtk::Image *icon, pa_proplist *l, const char *def) {
@@ -1078,12 +1157,15 @@ void MainWindow::reallyUpdateDeviceVisibility() {
     for (std::map<uint32_t, SinkInputWidget*>::iterator i = sinkInputWidgets.begin(); i != sinkInputWidgets.end(); ++i) {
         SinkInputWidget* w = i->second;
 
+        w->updating = true;
+        w->updateDeviceComboBox();
+
         if (sinkWidgets.size() > 1) {
             w->directionLabel->show();
-            w->deviceButton->show();
+            w->deviceComboBox->show();
         } else {
             w->directionLabel->hide();
-            w->deviceButton->hide();
+            w->deviceComboBox->hide();
         }
 
         if (showSinkInputType == SINK_INPUT_ALL || w->type == showSinkInputType) {
@@ -1091,6 +1173,8 @@ void MainWindow::reallyUpdateDeviceVisibility() {
             is_empty = false;
         } else
             w->hide();
+
+        w->updating = false;
     }
 
     if (eventRoleWidget)
@@ -1106,12 +1190,15 @@ void MainWindow::reallyUpdateDeviceVisibility() {
     for (std::map<uint32_t, SourceOutputWidget*>::iterator i = sourceOutputWidgets.begin(); i != sourceOutputWidgets.end(); ++i) {
         SourceOutputWidget* w = i->second;
 
+        w->updating = true;
+        w->updateDeviceComboBox();
+
         if (sourceWidgets.size() > 1) {
             w->directionLabel->show();
-            w->deviceButton->show();
+            w->deviceComboBox->show();
         } else {
             w->directionLabel->hide();
-            w->deviceButton->hide();
+            w->deviceComboBox->hide();
         }
 
         if (showSourceOutputType == SOURCE_OUTPUT_ALL || w->type == showSourceOutputType) {
@@ -1119,6 +1206,8 @@ void MainWindow::reallyUpdateDeviceVisibility() {
             is_empty = false;
         } else
             w->hide();
+
+        w->updating = false;
     }
 
     if (is_empty)
@@ -1241,18 +1330,18 @@ void MainWindow::removeClient(uint32_t index) {
 }
 
 void MainWindow::removeAllWidgets() {
-    for (std::map<uint32_t, SinkInputWidget*>::iterator it = sinkInputWidgets.begin(); it != sinkInputWidgets.end(); ++it)
-        removeSinkInput(it->first);
-    for (std::map<uint32_t, SourceOutputWidget*>::iterator it = sourceOutputWidgets.begin(); it != sourceOutputWidgets.end(); ++it)
-        removeSourceOutput(it->first);
-    for (std::map<uint32_t, SinkWidget*>::iterator it = sinkWidgets.begin(); it != sinkWidgets.end(); ++it)
-        removeSink(it->first);
-    for (std::map<uint32_t, SourceWidget*>::iterator it = sourceWidgets.begin(); it != sourceWidgets.end(); ++it)
-        removeSource(it->first);
-    for (std::map<uint32_t, CardWidget*>::iterator it = cardWidgets.begin(); it != cardWidgets.end(); ++it)
-       removeCard(it->first);
-    for (std::map<uint32_t, char*>::iterator it = clientNames.begin(); it != clientNames.end(); ++it)
-        removeClient(it->first);
+    while (!sinkInputWidgets.empty())
+        removeSinkInput(sinkInputWidgets.begin()->first);
+    while (!sourceOutputWidgets.empty())
+        removeSourceOutput(sourceOutputWidgets.begin()->first);
+    while (!sinkWidgets.empty())
+        removeSink(sinkWidgets.begin()->first);
+    while (!sourceWidgets.empty())
+        removeSource(sourceWidgets.begin()->first);
+    while (!cardWidgets.empty())
+        removeCard(cardWidgets.begin()->first);
+    while (!clientNames.empty())
+        removeClient(clientNames.begin()->first);
     deleteEventRoleWidget();
 }
 
