@@ -25,6 +25,9 @@
 #include <pulse/pulseaudio.h>
 #include <pulse/ext-stream-restore.h>
 #include <pulse/ext-device-manager.h>
+#ifdef HAVE_PULSE_MESSAGING_API
+#include <json-glib/json-glib.h>
+#endif
 
 #include <canberra-gtk.h>
 
@@ -40,6 +43,13 @@
 #include "sourceoutputwidget.h"
 #include "rolewidget.h"
 #include "mainwindow.h"
+#include "pavuapplication.h"
+
+#include <unordered_map>
+#include <utility>
+#include <memory>
+
+using WindowAndCardName = std::pair<MainWindow*, std::string>;
 
 static pa_context* context = NULL;
 static pa_mainloop_api* api = NULL;
@@ -56,7 +66,7 @@ void show_error(const char *txt) {
     Gtk::MessageDialog dialog(buf, false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_CLOSE, true);
     dialog.run();
 
-    Gtk::Main::quit();
+    PavuApplication::get_instance().quit();
 }
 
 static void dec_outstanding(MainWindow *w) {
@@ -69,7 +79,276 @@ static void dec_outstanding(MainWindow *w) {
     }
 }
 
-void card_cb(pa_context *, const pa_card_info *i, int eol, void *userdata) {
+#ifdef HAVE_PULSE_MESSAGING_API
+
+std::string card_message_handler_path(const std::string& name) {
+    return "/card/" + name;
+}
+
+std::string card_bluez_message_handler_path(const std::string& name) {
+    return "/card/" + name + "/bluez";
+}
+
+static void context_bluetooth_card_codec_list_cb(pa_context *c, int success, char *response, void *userdata) {
+    auto u = std::unique_ptr<WindowAndCardName>(reinterpret_cast<WindowAndCardName*>(userdata));
+
+    if (!success)
+        return;
+
+    int err = 0;
+
+    GError *gerror = NULL;
+    JsonParser *parser = json_parser_new();
+
+    if (!json_parser_load_from_data(parser, response, strlen(response), &gerror)) {
+        g_debug(_("could not read JSON from list-codecs message response: %s"), gerror->message);
+        g_error_free(gerror);
+        g_object_unref(parser);
+        return;
+    }
+
+    JsonNode *root = json_parser_get_root(parser);
+
+    if (!root || JSON_NODE_TYPE(root) != JSON_NODE_ARRAY) {
+        g_debug(_("list-codecs message response is not a JSON array"));
+        g_object_unref(parser);
+        return;
+    }
+
+    JsonArray *array = json_node_get_array(root);
+
+    std::unordered_map<std::string, std::string> codecs;
+
+    for (guint i = 0; i < json_array_get_length(array); ++i) {
+        const char *path;
+        const char *description;
+        JsonNode *v;
+
+        JsonObject *object = json_array_get_object_element(array, i);
+        if (!object) {
+            err = -1;
+            break;
+        }
+
+        v = json_object_get_member(object, "name");
+        if (!v) {
+            err = -1;
+            break;
+        }
+
+        path = json_node_get_string(v);
+        if (!path) {
+            err = -1;
+            break;
+        }
+
+        v = json_object_get_member(object, "description");
+        if (!v) {
+            err = -1;
+            break;
+        }
+
+        description = json_node_get_string(v);
+        if (!description)
+            description = "";
+
+        codecs[path] = description;
+    }
+
+    g_object_unref(parser);
+
+    if (err < 0) {
+        g_debug(_("list-codecs message response could not be parsed correctly"));
+        codecs.clear();
+        return;
+    }
+
+    u->first->updateCardCodecs(u->second, codecs);
+}
+
+static void context_bluetooth_card_active_codec_cb(pa_context *c, int success, char *response, void *userdata) {
+    auto u = std::unique_ptr<WindowAndCardName>(reinterpret_cast<WindowAndCardName*>(userdata));
+
+    if (!success)
+        return;
+
+    const char *name;
+    GError *gerror = NULL;
+
+    JsonParser *parser = json_parser_new();
+
+    if (!json_parser_load_from_data(parser, response, strlen(response), &gerror)) {
+        g_debug(_("could not read JSON from get-codec message response: %s"), gerror->message);
+        g_error_free(gerror);
+        g_object_unref(parser);
+        return;
+    }
+
+    JsonNode *root = json_parser_get_root(parser);
+
+    if (!root || JSON_NODE_TYPE(root) != JSON_NODE_VALUE) {
+        g_debug(_("get-codec message response is not a JSON value"));
+        g_object_unref(parser);
+        return;
+    }
+
+    name = json_node_get_string(root);
+
+    if (!name) {
+        g_debug(_("could not get codec name from get-codec message response"));
+        g_object_unref(parser);
+        return;
+    }
+
+    u->first->setActiveCodec(u->second, name);
+
+    g_object_unref(parser);
+}
+
+static void context_card_profile_is_sticky_cb(pa_context *c, int success, char *response, void *userdata) {
+    auto u = std::unique_ptr<WindowAndCardName>(reinterpret_cast<WindowAndCardName*>(userdata));
+
+    if (!success)
+        return;
+
+    gboolean profile_is_sticky;
+    GError *gerror = NULL;
+
+    JsonParser *parser = json_parser_new();
+
+    if (!json_parser_load_from_data(parser, response, strlen(response), &gerror)) {
+        g_debug(_("could not read JSON from get-profile-sticky message response: %s"), gerror->message);
+        g_error_free(gerror);
+        g_object_unref(parser);
+        return;
+    }
+
+    JsonNode *root = json_parser_get_root(parser);
+
+    if (!root || JSON_NODE_TYPE(root) != JSON_NODE_VALUE) {
+        g_debug(_("get-profile-sticky message response is not a JSON value"));
+        g_object_unref(parser);
+        return;
+    }
+
+    profile_is_sticky = json_node_get_boolean(root);
+
+    u->first->setCardProfileIsSticky(u->second, profile_is_sticky);
+
+    g_object_unref(parser);
+}
+
+template<typename U> void send_message(pa_context *c, const char *target, const char *request, pa_context_string_cb_t cb, const U& u)
+{
+    auto send_message_userdata = new U(u);
+
+    pa_operation *o = pa_context_send_message_to_object(c, target, request, NULL, cb, send_message_userdata);
+
+    if (!o) {
+        delete send_message_userdata;
+        g_debug(_("pa_context_send_message_to_object() failed: %s"), pa_strerror(pa_context_errno(context)));
+        return;
+    }
+    pa_operation_unref(o);
+}
+
+static void context_message_handlers_cb(pa_context *c, int success, char *response, void *userdata) {
+    auto u = std::unique_ptr<WindowAndCardName>(reinterpret_cast<WindowAndCardName*>(userdata));
+
+    if (!success)
+        return;
+
+    int err = 0;
+
+    GError *gerror = NULL;
+    JsonParser *parser = json_parser_new();
+
+    if (!json_parser_load_from_data(parser, response, strlen(response), &gerror)) {
+        g_debug(_("could not read JSON from list-handlers message response: %s"), gerror->message);
+        g_error_free(gerror);
+        g_object_unref(parser);
+        return;
+    }
+
+    JsonNode *root = json_parser_get_root(parser);
+
+    if (!root || JSON_NODE_TYPE(root) != JSON_NODE_ARRAY) {
+        g_debug(_("list-handlers message response is not a JSON array"));
+        g_object_unref(parser);
+        return;
+    }
+
+    JsonArray *array = json_node_get_array(root);
+
+    std::unordered_map<std::string, std::string> message_handler_map;
+
+    for (guint i = 0; i < json_array_get_length(array); ++i) {
+        const char *path;
+        const char *description;
+        JsonNode *v;
+
+        JsonObject *object = json_array_get_object_element(array, i);
+        if (!object) {
+            err = -1;
+            break;
+        }
+
+        v = json_object_get_member(object, "name");
+        if (!v) {
+            err = -1;
+            break;
+        }
+
+        path = json_node_get_string(v);
+        if (!path) {
+            err = -1;
+            break;
+        }
+
+        v = json_object_get_member(object, "description");
+        if (!v) {
+            err = -1;
+            break;
+        }
+
+        description = json_node_get_string(v);
+        if (!description)
+            description = "";
+
+        message_handler_map[path] = description;
+    }
+
+    g_object_unref(parser);
+
+    if (err < 0) {
+        g_debug(_("list-handlers message response could not be parsed correctly"));
+        message_handler_map.clear();
+        return;
+    }
+
+    /* only send requests if card bluez message handler is registered */
+    auto e = message_handler_map.find(card_bluez_message_handler_path(u->second));
+
+    if (e != message_handler_map.end()) {
+
+        /* get-codec: retrieve active codec name */
+        send_message(c, e->first.c_str(), "get-codec", context_bluetooth_card_active_codec_cb, *u);
+
+        /* list-codecs: retrieve list of codecs */
+        send_message(c, e->first.c_str(), "list-codecs", context_bluetooth_card_codec_list_cb, *u);
+    }
+
+    /* send requests to card if card message handler is registered */
+    e = message_handler_map.find(card_message_handler_path(u->second));
+
+    if (e != message_handler_map.end()) {
+        /* get-profile-sticky: retrieve active codec name */
+        send_message(c, e->first.c_str(), "get-profile-sticky", context_card_profile_is_sticky_cb, *u);
+    }
+}
+#endif
+
+void card_cb(pa_context *c, const pa_card_info *i, int eol, void *userdata) {
     MainWindow *w = static_cast<MainWindow*>(userdata);
 
     if (eol < 0) {
@@ -86,6 +365,11 @@ void card_cb(pa_context *, const pa_card_info *i, int eol, void *userdata) {
     }
 
     w->updateCard(*i);
+
+#ifdef HAVE_PULSE_MESSAGING_API
+    /* initiate requesting bluetooth codec list */
+    send_message(c, "/core", "list-handlers", context_message_handlers_cb, WindowAndCardName(w, i->name));
+#endif
 }
 
 #if HAVE_EXT_DEVICE_RESTORE_API
@@ -571,7 +855,7 @@ void context_state_callback(pa_context *c, void *userdata) {
 
         case PA_CONTEXT_TERMINATED:
         default:
-            Gtk::Main::quit();
+            w->get_application()->quit();
             return;
     }
 }
@@ -613,7 +897,7 @@ gboolean connect_to_pulse(gpointer userdata) {
         else {
             if(!retry) {
                 reconnect_timeout = -1;
-                Gtk::Main::quit();
+                w->get_application()->quit();
             } else {
                 g_debug(_("Connection failed, attempting reconnect"));
                 reconnect_timeout = 5;
